# coding=UTF-8
'''
Initial method designed in December 2012, School of Electrical Engineering, Univeristy of Belgrade
Python implementation added on 7 May 2015 at the University of Manchester, School of Computer Science
Updated on 11.16.2021 at Serbian AI Society
@author: Nikola Milosevic
'''
from typing import List

from nltk.tokenize import word_tokenize
import nltk
nltk.download('punkt')

rules = {
    'ovnicxki': '',
    'ovnicxka': '',
    'ovnika': '',
    'ovniku': '',
    'ovnicxe': '',
    'kujemo': '',
    'ovacyu': '',
    'ivacyu': '',
    'isacyu': '',
    'dosmo': '',
    'ujemo': '',
    'ijemo': '',
    'ovski': '',
    'ajucxi': '',
    'icizma': '',
    'ovima': '',
    'ovnik': '',
    'ognu': '',
    'inju': '',
    'enju': '',
    'cxicyu': '',
    'sxtva': '',
    'ivao': '',
    'ivala': '',
    'ivalo': '',
    'skog': '',
    'ucxit': '',
    'ujesx': '',
    'ucyesx': '',
    'ocyesx': '',
    'osmo': '',
    'ovao': '',
    'ovala': '',
    'ovali': '',
    'ismo': '',
    'ujem': '',
    'esmo': '',
    'asmo': '',  # pravi gresku kod pevasmo
    'zxemo': '',
    'cyemo': '',
    'cyemo': '',
    'bemo': '',
    'ovan': '',
    'ivan': '',
    'isan': '',
    'uvsxi': '',
    'ivsxi': '',
    'evsxi': '',
    'avsxi': '',
    'sxucyi': '',
    'uste': '',
    'icxe': 'i',  # bilo ik
    'acxe': 'ak',
    'uzxe': 'ug',
    'azxe': 'ag',  # mozda treba az, pokazati, pokazxe
    'aci': 'ak',
    'oste': '',
    'aca': '',
    'enu': '',
    'enom': '',
    'enima': '',
    'eta': '',
    'etu': '',
    'etom': '',
    'adi': '',
    'alja': '',
    'nju': 'nj',
    'lju': '',
    'lja': '',
    'lji': '',
    'lje': '',
    'ljom': '',
    'ljama': '',
    'zi': 'g',
    'etima': '',
    'ac': '',
    'becyi': 'beg',
    'nem': '',
    'nesx': '',
    'ne': '',
    'nemo': '',
    'nimo': '',
    'nite': '',
    'nete': '',
    'nu': '',
    'ce': '',
    'ci': '',
    'cu': '',
    'ca': '',
    'cem': '',
    'cima': '',
    'sxcyu': 's',
    'ara': 'r',
    'iste': '',
    'este': '',
    'aste': '',
    'ujte': '',
    'jete': '',
    'jemo': '',
    'jem': '',
    'jesx': '',
    'ijte': '',
    'inje': '',
    'anje': '',
    'acxki': '',
    'anje': '',
    'inja': '',
    'cima': '',
    'alja': '',
    'etu': '',
    'nog': '',
    'omu': '',
    'emu': '',
    'uju': '',
    'iju': '',
    'sko': '',
    'eju': '',
    'ahu': '',
    'ucyu': '',
    'icyu': '',
    'ecyu': '',
    'acyu': '',
    'ocu': '',
    'izi': 'ig',
    'ici': 'ik',
    'tko': 'd',
    'tka': 'd',
    'ast': '',
    'tit': '',
    'nusx': '',
    'cyesx': '',
    'cxno': '',
    'cxni': '',
    'cxna': '',
    'uto': '',
    'oro': '',
    'eno': '',
    'ano': '',
    'umo': '',
    'smo': '',
    'imo': '',
    'emo': '',
    'ulo': '',
    'sxlo': '',
    'slo': '',
    'ila': '',
    'ilo': '',
    'ski': '',
    'ska': '',
    'elo': '',
    'njo': '',
    'ovi': '',
    'evi': '',
    'uti': '',
    'iti': '',
    'eti': '',
    'ati': '',
    'vsxi': '',
    'vsxi': '',
    'ili': '',
    'eli': '',
    'ali': '',
    'uji': '',
    'nji': '',
    'ucyi': '',
    'sxcyi': '',
    'ecyi': '',
    'ucxi': '',
    'oci': '',
    'ove': '',
    'eve': '',
    'ute': '',
    'ste': '',
    'nte': '',
    'kte': '',
    'jte': '',
    'ite': '',
    'ete': '',
    'cyi': '',
    'usxe': '',
    'esxe': '',
    'asxe': '',
    'une': '',
    'ene': '',
    'ule': '',
    'ile': '',
    'ele': '',
    'ale': '',
    'uke': '',
    'tke': '',
    'ske': '',
    'uje': '',
    'tje': '',
    'ucye': '',
    'sxcye': '',
    'icye': '',
    'ecye': '',
    'ucxe': '',
    'oce': '',
    'ova': '',
    'eva': '',
    'ava': 'av',
    'uta': '',
    'ata': '',
    'ena': '',
    'ima': '',
    'ama': '',
    'ela': '',
    'ala': '',
    'aka': '',
    'aja': '',
    'jmo': '',
    'oga': '',
    'ega': '',
    'aća': '',
    'oca': '',
    'aba': '',
    'cxki': '',
    'ju': '',
    'hu': '',
    'cyu': '',
    'cu': '',
    'ut': '',
    'it': '',
    'et': '',
    'at': '',
    'usx': '',
    'isx': '',
    'esx': '',
    'esx': '',
    'uo': '',
    'no': '',
    'mo': '',
    'mo': '',
    'lo': '',
    'ko': '',
    'io': '',
    'eo': '',
    'ao': '',
    'un': '',
    'an': '',
    'om': '',
    'ni': '',
    'im': '',
    'em': '',
    'uk': '',
    'uj': '',
    'oj': '',
    'li': '',
    'ci': '',
    'uh': '',
    'oh': '',
    'ih': '',
    'eh': '',
    'ah': '',
    'og': '',
    'eg': '',
    'te': '',
    'sxe': '',
    'le': '',
    'ke': '',
    'ko': '',
    'ka': '',
    'ti': '',
    'he': '',
    'cye': '',
    'cxe': '',
    'ad': '',
    'ecy': '',
    'ac': '',
    'na': '',
    'ma': '',
    'ul': '',
    'ku': '',
    'la': '',
    'nj': 'nj',
    'lj': 'lj',
    'ha': '',
    'a': '',
    'e': '',
    'u': '',
    'sx': '',
    'o': '',
    'i': '',
    'j': '',
    'i': ''
}
dictionary = {
    # biti glagol
    'bih': 'biti',
    'bi': 'biti',
    'bismo': 'biti',
    'biste': 'biti',
    'bisxe': 'biti',
    'budem': 'biti',
    'budesx': 'biti',
    'bude': 'biti',
    'budemo': 'biti',
    'budete': 'biti',
    'budu': 'biti',
    'bio': 'biti',
    'bila': 'biti',
    'bili': 'biti',
    'bile': 'biti',
    'biti': 'biti',
    'bijah': 'biti',
    'bijasxe': 'biti',
    'bijasmo': 'biti',
    'bijaste': 'biti',
    'bijahu': 'biti',
    'besxe': 'biti',
    # jesam
    'sam': 'jesam',
    'si': 'jesam',
    'je': 'jesam',
    'smo': 'jesam',
    'ste': 'jesam',
    'su': 'jesam',
    'jesam': 'jesam',
    'jesi': 'jesam',
    'jeste': 'jesam',
    'jesmo': 'jesam',
    'jeste': 'jesam',
    'jesu': 'jesam',
    # glagol hteti
    'cyu': 'hteti',
    'cyesx': 'hteti',
    'cye': 'hteti',
    'cyemo': 'hteti',
    'cyete': 'hteti',
    'hocyu': 'hteti',
    'hocyesx': 'hteti',
    'hocye': 'hteti',
    'hocyemo': 'hteti',
    'hocyete': 'hteti',
    'hocye': 'hteti',
    'hteo': 'hteti',
    'htela': 'hteti',
    'hteli': 'hteti',
    'htelo': 'hteti',
    'htele': 'hteti',
    'htedoh': 'hteti',
    'htede': 'hteti',
    'htede': 'hteti',
    'htedosmo': 'hteti',
    'htedoste': 'hteti',
    'htedosxe': 'hteti',
    'hteh': 'hteti',
    'hteti': 'hteti',
    'htejucyi': 'hteti',
    'htevsxi': 'hteti',
    # glagol moći
    'mogu': 'mocyi',
    'možeš': 'mocyi',
    'može': 'mocyi',
    'možemo': 'mocyi',
    'možete': 'mocyi',
    'mogao': 'mocyi',
    'mogli': 'mocyi',
    'moći': 'mocyi'
}


def stem_arr(text: str) -> List[str]:
    """
    
    :param text:
    :return:
    """
    text = text.lower()
    text = text.replace("š", "sx")
    text = text.replace("č", "cx")
    text = text.replace("ć", "cy")
    text = text.replace("đ", "dx")
    text = text.replace("ž", "zx")
    lam = word_tokenize(text)
    i = 0
    for word in lam:
        for key in dictionary:
            if key == word:
                lam[i] = dictionary[key]
                break
        for key in rules:
            if (word.endswith(key) and len(word) - len(key) > 2):
                lam[i] = word[:-len(key)] + rules[key]
                break
        i = i + 1
    return lam


def stem_str(text: str) -> str:
    '''

    :param text:
    :return:
    '''
    text = text.lower()
    text = text.replace("š", "sx")
    text = text.replace("č", "cx")
    text = text.replace("ć", "cy")
    text = text.replace("đ", "dx")
    text = text.replace("ž", "zx")
    text = text.replace("Š", "sx")
    text = text.replace("Č", "cx")
    text = text.replace("Ć", "cy")
    text = text.replace("Đ", "dx")
    text = text.replace("Ž", "zx")
    text = text.replace("“", "\"")
    text = text.replace("\"", "")
    text = text.replace("”", "\"")
    text = text.replace("'", "\"")
    text = text.replace("’", "\"")
    text = text.replace("€", "eur")
    text = text.replace("„", "\"")
    lam = word_tokenize(text)
    i = 0
    for word in lam:
        for key in dictionary:
            if key == word:
                lam[i] = dictionary[key]
                break
        for key in rules:
            # Tokenize only words larger than 2 characters, apart from modal verbs
            if (word.endswith(key) and len(word) - len(key) > 2):
                lam[i] = word[:-len(key)] + rules[key]
                break
        i = i + 1
    end_str = ""
    for word in lam:
        end_str = end_str + " " + word
    return end_str



import os

import numpy as np
from collections import defaultdict
import time


class SpellCheck:
    def __init__(self, dictionary):
        if dictionary in ['sr-latin','sr-cirilic']:
            path = os.path.abspath(__file__).split(os.path.sep)
            path_len = len(path)
            i = 0
            new_path = ""
            for p in path:
                if i<path_len-2:
                    new_path = new_path + p + os.path.sep
                    i = i + 1
            if dictionary == 'sr-latin':
                dictionary = new_path + os.path.sep+"Resursi"+os.path.sep+"Recnici"+os.path.sep+ "Serbian (Latin).dic"
            if dictionary == 'sr-cirilic':
                dictionary = new_path + "/Resursi/Recnici/Serbian (Cyrilic).dic"

        self.trie = {}
        self.max_cost = 2
        self.load_dictionary(dictionary)


    def load_dictionary(self, dictionary):
        with open(dictionary, 'r') as f:
            for line in f:
                self.add_word(line.strip())

    def add_word(self, word):
        node = self.trie
        for char in word:
            if char not in node:
                node[char] = {}
            node = node[char]
        node['$'] = word

    def search_trie(self, node, word, cost, results):
        if '$' in node:
            results.append((node['$'], cost))
        if not word:
            return
        if cost > self.max_cost:
            return
        for c in node:
            if c == '$':
                continue
            if c == word[0]:
                self.search_trie(node[c], word[1:], cost, results)
            else:
                self.search_trie(node[c], word[1:], cost + 1, results)
                self.search_trie(node[c], word, cost + 1, results)
                self.search_trie(node[c], word[1:], cost + 1, results)
    def lev_distance(self,s, t):
        m, n = len(s), len(t)
        d = np.zeros((m+1, n+1), dtype=int)
        for i in range(m+1):
            d[i,0] = i
        for j in range(n+1):
            d[0,j] = j
        for j in range(1,n+1):
            for i in range(1,m+1):
                if s[i-1] == t[j-1]:
                    cost = 0
                else:
                    cost = 1
                d[i,j] = min(d[i-1,j]+1, d[i,j-1]+1, d[i-1,j-1]+cost)
        return d[m,n]

    def find_closest(self,word,candidates):
        min_distance = float('inf')
        closest_word = None
        for dict_word1 in candidates:
            dict_word = dict_word1
            distance = self.lev_distance(word, dict_word)
            if distance < min_distance:
                min_distance = distance
                closest_word = dict_word
        return closest_word


    def spellcheck(self, query):
        results = []
        self.search_trie(self.trie, query, 0, results)
        results = list(set(val[0] for val in results))
        result = self.find_closest(query,results)

        return result
# Example usage:





from typing import Tuple, Dict, Callable, List
from operator import itemgetter

import numpy as np
import matplotlib.pyplot as plt
import tqdm

from globals import ModelConfig
from text_preprocessing import word_to_vector, make_context_matrix


def relu(tensor: np.ndarray) -> np.ndarray:
    """
    Does the relu function.
    f(x) = max(0, x)
    """
    return tensor * (tensor > 0.0)


def relu_derivative(tensor: np.ndarray) -> np.ndarray:
    """
    Does the derivative of relu
    """
    return 1.0 * (tensor > 0.0)


def tanh(tensor: np.ndarray) -> np.ndarray:
    """
    Does the hyperbolic tan elementwise on tensor
    """
    return np.tanh(tensor)


def tanh_derivative(tensor: np.ndarray) -> np.ndarray:
    """
    Does the derivative of tanh
    """
    return 1 - tanh(tensor) ** 2


def identity(tensor: np.ndarray) -> np.ndarray:
    """
    Does the y = x function on a tensor
    """
    return tensor


def identity_derivative(tensor: np.ndarray) -> np.ndarray:
    """
    Does the derivative of the identity function
    """
    return np.ones_like(tensor)


def softmax(tensor: np.ndarray) -> np.ndarray:
    """
    Does the safe softmax function with numerical stability
    """
    mxs = np.max(tensor, axis=1).reshape(-1, 1)
    exps = np.exp(tensor - mxs)

    sums = np.sum(exps, axis=1).reshape(-1, 1)

    return exps / sums


def make_network_hyperparameters(
    input_size: int, vector_space_size: int, output_size: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Initializes random weights and biases for the network
    Returns a pair of weights, input-hidden and hidden-output
    """
    input_2_hidden = np.random.uniform(-1, 1, size=(input_size, vector_space_size))
    hidden_2_output = np.random.uniform(-1, 1, size=(vector_space_size, output_size))
    return input_2_hidden, hidden_2_output


def forward_one_word(
    word_vector: np.ndarray,
    network_hyperparams: Tuple[np.ndarray, np.ndarray],
    activation: Callable[[np.ndarray], np.ndarray],
) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
    """
    Does the forward pass for one word of the network.
    Network hyperparameters are in a tuple of input to hidden and hidden to output weights

    Returns a tuple with all the transfer and activation pairs (in order of layers),
    with the last layers' activation being the output
    """
    input_2_hidden_weights, hidden_2_output_weights = network_hyperparams

    input_2_hidden_transfer = word_vector @ input_2_hidden_weights
    input_2_hidden_activation = activation(input_2_hidden_transfer)

    hidden_2_output_transfer = input_2_hidden_activation @ hidden_2_output_weights
    hidden_2_output_activation = softmax(hidden_2_output_transfer)

    return (input_2_hidden_transfer, input_2_hidden_activation), (
        hidden_2_output_transfer,
        hidden_2_output_activation,
    )


def backward_one_word(
    word_vector: np.ndarray,
    context_matrix: np.ndarray,
    forward_results: Tuple[
        Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]
    ],
    network_hyperparams: Tuple[np.ndarray, np.ndarray],
    activation_derivative: Callable[[np.ndarray], np.ndarray],
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Takes the word vector it's working on currently, as well as its context.
    Also, the results for all steps in the forward pass and the weights for each layer

    Returns gradients for the network hyperparameters in the order of layers
    """
    _, hidden_2_output_weights = network_hyperparams
    (input_2_hidden_transfer, input_2_hidden_activation), (
        _,
        hidden_2_output_activation,
    ) = forward_results

    grad_hidden_2_output_transfer = (
        np.sum(hidden_2_output_activation - context_matrix, axis=0, keepdims=True)
        / context_matrix.shape[0]
    )
    grad_hidden_2_output_weights = (
        input_2_hidden_activation.T @ grad_hidden_2_output_transfer
    )

    grad_input_2_hidden_activation = (
        grad_hidden_2_output_transfer @ hidden_2_output_weights.T
    )
    grad_input_2_hidden_transfer = (
        grad_input_2_hidden_activation * activation_derivative(input_2_hidden_transfer)
    )
    grad_input_2_hidden_weights = word_vector.T @ grad_input_2_hidden_transfer

    return grad_input_2_hidden_weights, grad_hidden_2_output_weights


def loss_one_word(word_output: np.ndarray, context_matrix: np.ndarray) -> float:
    """
    Calculates log loss for a single word based on its output and context
    """
    loss = np.sum(-1 * (context_matrix * np.log(word_output)))
    return loss / context_matrix.shape[0]


def update_hyperparameter(
    hyperparameter: np.ndarray,
    grad_hyperparameter: np.ndarray,
    m: np.ndarray,
    v: np.ndarray,
    optimizer_params: Dict[str, float],
    learn_rate: float,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Does ADAM optimization on hyperparameter

    Returns the newly calculated hyperparameter and optimizer parameters
    """
    new_m = (
        optimizer_params["omega1"] * m
        + (1 - optimizer_params["omega1"]) * grad_hyperparameter
    )
    new_v = optimizer_params["omega2"] * v + (1 - optimizer_params["omega2"]) * (
        grad_hyperparameter**2
    )

    m_hat = new_m / (1 - optimizer_params["omega1"])
    v_hat = np.abs(new_v) / (1 - optimizer_params["omega2"])

    new_hyperparameter = (
        hyperparameter
        - (learn_rate / (v_hat + optimizer_params["norm"]) ** 0.5) * m_hat
    )
    return new_hyperparameter, new_m, new_v


def one_word_iteration(
    word_vector: np.ndarray,
    context_matrix: np.ndarray,
    network_hyperparameters: Tuple[np.ndarray, np.ndarray],
    ms: Tuple[np.ndarray, np.ndarray],
    vs: Tuple[np.ndarray, np.ndarray],
    optimizer_params: Dict[str, float],
    learn_rate: float,
) -> Tuple[
    Tuple[np.ndarray, np.ndarray],
    float,
    Tuple[np.ndarray, np.ndarray],
    Tuple[np.ndarray, np.ndarray],
]:
    """
    Does forward pass, loss calculation, backward pass and hyperparameter optimization for one word

    Returns the new hyperparameters and loss for that word, as well as the newly calculated optimizer parameters
    """
    forward_results = forward_one_word(word_vector, network_hyperparameters, tanh)

    _, (_, hidden_2_output_activation) = forward_results
    loss = loss_one_word(hidden_2_output_activation, context_matrix)

    grad_hyperparameters = backward_one_word(
        word_vector,
        context_matrix,
        forward_results,
        network_hyperparameters,
        tanh_derivative,
    )

    input_2_hidden_updated, hidden_2_output_updated = tuple(
        update_hyperparameter(
            hyperparameter, grad_hyperparameter, m, v, optimizer_params, learn_rate
        )
        for hyperparameter, grad_hyperparameter, m, v in zip(
            network_hyperparameters, grad_hyperparameters, ms, vs
        )
    )

    return (
        (input_2_hidden_updated[0], hidden_2_output_updated[0]),
        loss,
        (input_2_hidden_updated[1], hidden_2_output_updated[1]),
        (input_2_hidden_updated[2], hidden_2_output_updated[2]),
    )


def train(
    word_map: Dict[str, Dict[str, List[int]]],
    ngram_vector_map: Dict[str, int],
    word_vector_map: Dict[str, int],
    model_config: ModelConfig,
    plot_graph: bool = False,
) -> np.ndarray:
    """
    Given a word map with input vectors and their context vectors, embeds the vector space.
    plot_graph - whether to plot the loss graph
    """
    words = [*word_map.keys()]
    hyperparameters = make_network_hyperparameters(
        len(ngram_vector_map),
        model_config.vector_space_size,
        len(word_vector_map),
    )

    total_loss = []
    for _ in tqdm.tqdm(range(model_config.epochs)):
        epoch_loss = 0.0
        ms = vs = (0.0, 0.0)

        np.random.shuffle(words)
        for word in words:
            word_dict = itemgetter(word)(word_map)
            word_vector = word_to_vector(word_dict["input"], ngram_vector_map)
            word_context = make_context_matrix(word_dict["context"], word_vector_map)

            hyperparameters, word_loss, ms, vs = one_word_iteration(
                word_vector, word_context, hyperparameters, ms, vs, model_config.optimizer, model_config.learn_rate
            )
            epoch_loss += word_loss

        total_loss.append(epoch_loss)

    if plot_graph:
        plt.plot(total_loss)
        plt.show()

    return hyperparameters[0]



import classla

classla.download('sr')
nlp = classla.Pipeline('sr', processors='tokenize,pos,lemma')

def get_lemmas(text: str):
    doc = nlp(text)
    dc = doc.to_dict()
    lemmas = []
    for da in dc:
        arr = da[0]
        print(arr)
        for token in arr:
            lemmas.append(token['lemma'])
    return lemmas

lemme = get_lemmas("Milica voli da ide u školu koja se nalazi u Novom Sadu. Nije ništa rekla. ")
print(lemme)



import re
import string
from typing import List, Dict, Tuple
from operator import itemgetter

import numpy as np

from globals import ModelConfig, LanguageConfig
from io_utils import open_corpus


def remove_white_space(text: str) -> str:
    """
    Removes multiple occurrences of space, newlines, return lines with a single space
    """
    cleaned_text = re.sub(r"(\s+|\n+|\r+)", " ", text)
    return cleaned_text


def remove_numbers(text: str) -> str:
    """
    Removes all numbers in text
    """
    cleaned_text = re.sub(r"[0-9]+", " ", text)
    return cleaned_text


def remove_punctuation(text: str) -> str:
    """
    Removes all occurrences of punctuation
    """
    cleaned_text = text.translate(str.maketrans("", "", string.punctuation))
    return cleaned_text


def to_lowercase(text: str) -> str:
    """
    Turns all words into lowercase
    """
    return text.lower()


def clean_text(text: str) -> str:
    """
    Cleans text using above functions
    """
    no_whitespace = remove_white_space(text)
    no_numbers = remove_numbers(no_whitespace)
    no_punctuation = remove_punctuation(no_numbers)
    lowercase = to_lowercase(no_punctuation)
    return lowercase


def make_word_list(text: str) -> List[str]:
    """
    Makes a list from single whitespace separated text
    """
    split_text = text.split(" ")
    return [word for word in split_text if word != ""]


def remove_stopwords(word_list: List[str], stopwords_path: str) -> List[str]:
    """
    Removes all stopwords from text.
    Stopwords path is the path to the comma separated stopwords.
    """
    # preprocess stopwords
    with open(stopwords_path, "r", encoding="utf-8") as f:
        stopwords = f.read()
    stopwords = stopwords.split(",")

    word_list = [word for word in word_list if word not in stopwords]
    return word_list


def make_ngram_map(word_list: List[str], ngram_size: int) -> Dict[str, List[str]]:
    """
    Makes a map of all the words and their ngrams given a list of words.
    Inside the list of ngrams is also the word itself as a ngram (as per facebook docs)
    """
    ngram_map = {}
    for word in word_list:
        ngram_map[word] = [
            word[i : i + ngram_size] for i in range(0, len(word) - (ngram_size - 1))
        ]
        ngram_map[word].append(word)
        word_list = [other_word for other_word in word_list if other_word != word]

    return ngram_map


def make_context_map(
    word_list: List[str],
    left_context_window: int,
    right_context_window: int,
    self_context: bool,
) -> Dict[str, List[str]]:
    """
    Makes a map of the context of a single word given how far left and right the context stretches.
    The word itself is inside its own context.
    """
    context_map = {}

    for i, word in enumerate(word_list):
        temp_left_context_window = left_context_window if i > left_context_window else i

        if word not in context_map:
            context_map[word] = word_list[
                i - temp_left_context_window : i + right_context_window + 1
            ]
        else:
            context_map[word].extend(
                word_list[i - temp_left_context_window : i + right_context_window + 1]
            )

        if not self_context:
            context_map[word].remove(word)

        # removes duplicates, but could be slow
        context_map[word] = list(set(context_map[word]))

    return context_map


def make_ngram_vector_map(ngram_map: Dict[str, List[str]]) -> Dict[str, int]:
    """
    Makes a vector for each ngram and maps the ngram to that vector.
    """
    # python magic. list of lists -> list of all elements using iterators
    all_ngrams = [ngram for ngram_list in [*ngram_map.values()] for ngram in ngram_list]
    unique_ngrams = np.unique(all_ngrams)

    ngram_vector_map = {}
    for i, ngram in enumerate(unique_ngrams):
        ngram_vector_map[ngram] = i

    return ngram_vector_map


def make_word_vector_map(word_list: List[str]) -> Dict[str, int]:
    """
    Given the list of words, one hot encodes them all
    """
    unique_words = np.unique(word_list)

    word_vector_map = {}
    for i, word in enumerate(unique_words):
        word_vector_map[word] = i

    return word_vector_map


def _spawn_one_hot_encoded_vector(vector_size: int, one_index: int) -> np.ndarray:
    """
    Given the size of the vector and the index of the 1 returns a one hot encoded vector
    """
    ohe_vec = np.zeros(shape=(1, vector_size))
    ohe_vec[0, one_index] = 1.0
    return ohe_vec


def word_to_vector(
    ngram_list: List[int], ngram_vector_map: Dict[str, int]
) -> np.ndarray:
    """
    Given a list of ngrams of a word, returns the vector representing that word based on ngrams
    """
    vector_size = len(ngram_vector_map)
    # more python magic. itemgetter gets a collection of items from an iterable/collection all using iterables
    return np.sum(
        [_spawn_one_hot_encoded_vector(vector_size, ngram) for ngram in ngram_list],
        axis=0,
    ).reshape(1, -1)


def make_context_matrix(
    context_list: List[int], word_vector_map: Dict[str, int]
) -> np.ndarray:
    """
    Creates context matrix for given a list of the words context words
    """
    vector_size = len(word_vector_map)
    return np.vstack(
        [_spawn_one_hot_encoded_vector(vector_size, word) for word in context_list]
    )


def make_word_map(
    word_list: List[str],
    ngram_map: Dict[str, List[str]],
    context_map: Dict[str, List[str]],
    ngram_vectors: Dict[str, int],
    word_vector_map: Dict[str, int],
) -> Dict[str, Dict[str, List[int]]]:
    """
    Maps the input vector as well as the target matrix to their respective words

    Return dict in the shape of: {word: {input: vector, context: matrix}}
    """
    word_map = {}
    for word in np.unique(word_list):
        single_entry = {}
        ngrams = ngram_map[word]
        context = context_map[word]

        input_vector = (
            list(itemgetter(*ngrams)(ngram_vectors))
            if len(ngrams) > 1
            else [ngram_vectors[ngrams[0]]]
        )
        target_matrix = (
            list(itemgetter(*context)(word_vector_map))
            if len(context) > 1
            else [word_vector_map[context[0]]]
        )

        single_entry["input"] = input_vector
        single_entry["context"] = target_matrix

        word_map[word] = single_entry

    return word_map


def make_maps(
    word_list: List[str], model_config: ModelConfig
) -> Tuple[
    Dict[str, List[str]],
    Dict[str, List[str]],
    Dict[str, int],
    Dict[str, int],
    Dict[str, Dict[str, List[int]]],
]:
    """
    Makes all the needed maps for later use in the network.

    Returns ngram, context, ngram_vector, word_vector and word maps
    """
    ngram_map = make_ngram_map(word_list, model_config.ngram_size)
    context_map = make_context_map(
        word_list,
        left_context_window=model_config.context_window["left"],
        right_context_window=model_config.context_window["right"],
        self_context=model_config.self_context,
    )
    ngram_vectors = make_ngram_vector_map(ngram_map)
    word_vectors = make_word_vector_map(word_list)
    word_map = make_word_map(
        word_list, ngram_map, context_map, ngram_vectors, word_vectors
    )

    return ngram_map, context_map, ngram_vectors, word_vectors, word_map


def preprocess_corpus(
    language_config: LanguageConfig, model_config: ModelConfig
) -> Tuple[
    Dict[str, List[str]],
    Dict[str, List[str]],
    Dict[str, int],
    Dict[str, int],
    Dict[str, Dict[str, List[int]]],
]:
    """
    Given the path to the corpus, create dataset that will later be used in the NN
    """
    text = open_corpus(language_config.corpus_path)
    cleaned = clean_text(text)
    word_list = make_word_list(cleaned)
    word_list = remove_stopwords(word_list, language_config.stopwords_path)
    maps = make_maps(word_list, model_config)

    return maps



import os

from nltk import word_tokenize
from nltk.tag.hunpos import HunposTagger

from src import srbai


class POS_Tagger():
    def __init__(self):
        path = os.path.abspath(srbai.__file__).replace(os.path.sep + "__init__.py", "")
        os.environ["HUNPOS_TAGGER"] = os.environ["HUNPOS"] = path + os.path.sep+"Resursi"+os.path.sep+"PomocneDatoteke"+os.path.sep+"hunpos-tag.exe"
        print(os.environ["HUNPOS_TAGGER"])
        self.ht = HunposTagger(path + os.path.sep+'Resursi'+os.path.sep+'Modeli'+os.path.sep+'model.hunpos.mte5.defnpout',encoding='UTF-8')


    def tag(self, text):
        """
        Oznacava vrste reci u tekstu. Prvo slovo oznacava generalno vrstu reci:
        http://nlp.ffzg.hr/data/tagging/msd-hr.html
        - N - Imenica
        - V - Glagol
        - A - Pridev
        - P - Zamenice
        - S - Predlozi
        - Z - Znak interpunkcije
        - C - Veznik
        - M - Broj
        - R - Prilozi,
        - I - Uzvici
        - Q - Rečice
        - Y - Skraćenice
        - X - Ostalo
        :param text:
        :return:
        """
        tagged = self.ht.tag(word_tokenize(text))
        return tagged



from typing import Dict, Tuple, List

import numpy as np

from embedding import embed_word


def vector_cosine_similarity(vector1: np.ndarray, vector2: np.ndarray) -> float:
    """
    Does cosine similarity (from formula of vector product)
    Return interval is -1 to 1 with 1 being same and -1 being completely different
    """
    similarity = (
        vector1 @ vector2.T / (np.linalg.norm(vector1) * np.linalg.norm(vector2))
    )

    # due to mat mul the result is a shape(1, 1) np array and not a float, so indexing is needed
    return similarity[0, 0]


def word_similarity(
    word: str,
    word_map: Dict[str, Dict[str, List[int]]],
    ngram_vectors: Dict[str, int],
    vector_space: np.ndarray,
    ngram_size: int,
) -> Tuple[str, Dict[str, float]]:
    """
    Finds the cosine similarity between the word and the whole known dictionary in the word_map.
    Returns a dictionary with the key being the word and the value being the similarity between
    the queried word and the word in the key
    """
    embedded_word = embed_word(word, ngram_vectors, vector_space, ngram_size)

    if isinstance(embedded_word, int):
        raise Exception(f'Word "{word}" can\'t be embedded')

    similarities = {
        dict_word: vector_cosine_similarity(
            embedded_word, embed_word(dict_word, ngram_vectors, vector_space, ngram_size)
        )
        for dict_word in word_map.keys()
    }

    return word, similarities



import classla


class NER_classla():
    def __init__(self):
        classla.download('sr')
        self.nlp = classla.Pipeline('sr')

    def perform_NER(self, text):
        doc = self.nlp(text)
        result = []
        for ent in doc.to_dict()[0][0]:
            result.append({'text': ent["text"], "NER": ent["ner"], "dep_rel": ent["deprel"]})
        return result



import pickle
from typing import Any


def serialize_object(obj: Any, path: str) -> None:
    """
    Serializes object first to a pickle string then encodes that string into base64.
    """
    with open(path, "wb") as f:
        pickle.dump(obj, f)


def deserialize_object(path: str) -> Any:
    """
    Deserializes object encoded in base64 made from a pickle string.
    Assumes path exists.
    """
    with open(path, "rb") as f:
        obj = pickle.load(f)

    return obj


def open_corpus(path: str) -> str:
    """
    Opens a corpus and loads its content into a string.
    Assumes path exists.
    """
    with open(path, "r", encoding="utf-8") as f:
        content = f.read()

    return content



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="SrbAI",
    version="0.0.5",
    author="Serbian AI Society",
    author_email="nikola.milosevic86@gmail.com",
    description="Library for processing serbian language",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Serbian-AI-Society/SrbAI",
    project_urls={
        "Bug Tracker": "https://github.com/Serbian-AI-Society/SrbAI/issues",
        "Road Map":"https://github.com/Serbian-AI-Society/SrbAI/projects/1"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "nltk",
        "torch>=1.10.2",
        "classla==1.1.0",
        "nlu"

    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6.15",
)


from model import train
from text_preprocessing import preprocess_corpus
from globals import GlobalConfig
from io_utils import serialize_object, deserialize_object
from similarity import word_similarity
from embedding import embed_word, embed_word_map


def model_training():
    """
    Example function showing how to train the model for a given language
    """
    # Step 1. Load a configuration with which you would like to train the model
    # To load a config just type the folder name of the model and language
    cfg = GlobalConfig("5gram_5sctx_300vs", "serbian")

    # Step 2. Preprocess the corpus and prepare it for the NN
    _, _, ngram_vectors, word_vectors, word_map = preprocess_corpus(
        cfg.language_config, cfg.model_config
    )

    # Optional Step. Serialize the ngram_vectors and word_map so that they can be used later
    # Keep in mind that the save_dir from the config must exist, the serialization won't check for that itself
    serialize_object(ngram_vectors, cfg.language_config.get_ngram_vectors_file_path())
    serialize_object(word_map, cfg.language_config.get_word_map_file_path())

    # Step 3. Train the NN
    # The result of training will be the weights that lead to the embedding layer
    # If plot graph is true, it will plot the loss graph as well
    vector_space = train(word_map, ngram_vectors, word_vectors, cfg.model_config, plot_graph=False)

    # Optional Step. Save the weights that lead to the embedding layer
    serialize_object(vector_space, cfg.language_config.get_vector_space_file_path())


def similarity_for_word():
    """
    Example function that shows how finding most similar words for a word works
    """
    # Step 1. Load configs for both the model and language
    cfg = GlobalConfig("5gram_5sctx_300vs", "serbian")

    # Step 2. Load the ngram vectors, vector space and word map files
    ngram_vectors = deserialize_object(cfg.language_config.get_ngram_vectors_file_path())
    word_map = deserialize_object(cfg.language_config.get_word_map_file_path())
    vector_space = deserialize_object(cfg.language_config.get_vector_space_file_path())

    # Step 3. Run the word similarity function from the similarity module
    # It will return a tuple of the word itself and
    # a dictionary of words from the word map and their similarities w.r.t the passed word
    word, similarities = word_similarity("ekonom", word_map, ngram_vectors, vector_space, cfg.model_config.ngram_size)

    # Now you can do whatever you like with these similarities
    # Here are the top 5 similarities sorted in descending order
    print(sorted(similarities.items(), key=lambda item: item[1], reverse=True)[:5])


def embedding_things():
    """
    Example function of how to use the embedding module to embed a word or the whole current word map
    """
    # Step 1. Load the configurations for the model and language
    cfg = GlobalConfig("5gram_5sctx_300vs", "serbian")

    # Step 2. Load the ngram vectors, vector space and word map files
    ngram_vectors = deserialize_object(cfg.language_config.get_ngram_vectors_file_path())
    word_map = deserialize_object(cfg.language_config.get_word_map_file_path())
    vector_space = deserialize_object(cfg.language_config.get_vector_space_file_path())

    # Step 3.1 Use the embed word function to embed a single word into the vector space
    # This function will return only the vector
    vector = embed_word("ekonom", ngram_vectors, vector_space, cfg.model_config.ngram_size)
    print(vector)

    # Step 3.2 Use embed_word_map function to embed all the current words in the word map
    # This function will return a dictionary of all the words and their embedded vector representations
    embedded_word_map = embed_word_map(word_map, ngram_vectors, vector_space, cfg.model_config.ngram_size)
    # Print only 5 key-value pairs
    print(list(embedded_word_map.items())[:5])


if __name__ == "__main__":
    embedding_things()



from typing import List, Dict, Union

import numpy as np

from text_preprocessing import word_to_vector


def _iter_len(str_len: int, ngrams: int) -> int:
    return str_len - (ngrams - 1) if str_len > ngrams else 1


def make_ngram_list_from_word(word: str, ngram_size: int) -> List[str]:
    """
    Returns a list of ngrams as well as the word itself
    """
    return [
        word[i:i + ngram_size] for i in range(0, _iter_len(len(word), ngram_size))
    ]


def count_missing_ngrams(
    word_ngram_list: List[str], ngram_vectors: Dict[str, int]
) -> int:
    """
    Checks to see how many ngrams from a word are not in the dictionary of known ngrams from the preprocessed corpus
    """
    all_ngrams = ngram_vectors.keys()
    return len(word_ngram_list) - sum(
        [ngram in all_ngrams for ngram in word_ngram_list]
    )


def remove_missing_ngrams(
    word_ngram_list: List[str], ngram_vectors: Dict[str, int]
) -> List[str]:
    """
    Removes the ngrams not found in the original ngram dictionary
    """
    all_ngrams = ngram_vectors.keys()
    return [ngram for ngram in word_ngram_list if ngram in all_ngrams]


def word_vector(
    word: str, ngram_vectors: Dict[str, int], ngram_size: int
) -> Union[np.ndarray, int]:
    """
    Returns the vector representing the word

    If there are more than 2 unknown ngrams then it returns -1.
    2 was chosen because that is the minimum case for a new word that only has one ngram changed compared to its closest
    word in the dictionary
    """
    word_ngram_list = make_ngram_list_from_word(word, ngram_size)
    missing = count_missing_ngrams(word_ngram_list, ngram_vectors)

    if (
        (missing > 2)
        or (len(word_ngram_list) == 1 and missing == 1)
        or (len(word_ngram_list) == 2 and missing == 2)
    ):
        return -1

    if missing <= 2:
        word_ngram_list = remove_missing_ngrams(word_ngram_list, ngram_vectors)

    return word_to_vector(
        [ngram_vectors[ngram] for ngram in word_ngram_list], ngram_vectors
    )


def embed_word(
    word: str,
    ngram_vectors: Dict[str, int],
    vector_space: np.ndarray,
    ngram_size: int,
) -> Union[np.ndarray, int]:
    """
    Returns the embedded representation of the word in the vector space or -1 if the word can't be vectorized
    """
    word_vectorized = word_vector(word, ngram_vectors, ngram_size)

    if isinstance(word_vectorized, int):
        return -1

    embedded = word_vectorized @ vector_space
    return embedded


def embed_word_map(
    word_map: Dict[str, Dict[str, List[int]]],
    ngram_vectors: Dict[str, int],
    vector_space: np.ndarray,
    ngram_size: int,
) -> Dict[str, np.ndarray]:
    """
    Embeds the word map made in text preprocessing.

    Returns a dict where the keys are the words and values are their embedded vectors.
    """
    return {
        word: embed_word(word, ngram_vectors, vector_space, ngram_size)
        for word in word_map.keys()
    }



import os
from typing import Dict, Any

import yaml


class Config:
    def __init__(self) -> None:
        self.config_dir: str = os.path.join(".", "configs")

    def load_yaml(self, yaml_path: str) -> Dict[str, Any]:
        with open(yaml_path, "rb") as f:
            yaml_data = yaml.safe_load(f)

        return yaml_data


class ModelConfig(Config):
    def __init__(self, model_name: str) -> None:
        super().__init__()

        _yaml_data = self.load_yaml(
            os.path.join(self.config_dir, "models", model_name, "config.yaml")
        )

        _text_preproc = _yaml_data["text_preprocessing"]
        self.ngram_size: int = _text_preproc["ngram_size"]
        self.context_window: Dict[str, int] = _text_preproc["context_window"]
        self.self_context: bool = _text_preproc["self_context"]

        _neural_network = _yaml_data["neural_network"]
        self.vector_space_size: int = _neural_network["vector_space_size"]
        self.optimizer: Dict[str, float] = _neural_network["optimizer"]
        self.learn_rate: float = _neural_network["learn_rate"]
        self.epochs: int = _neural_network["epochs"]


class LanguageConfig(Config):
    def __init__(self, language: str) -> None:
        super().__init__()

        _yaml_data = self.load_yaml(
            os.path.join(self.config_dir, "languages", language, "config.yaml")
        )

        self.corpus_path: str = _yaml_data["corpus_path"]
        self.stopwords_path: str = _yaml_data["stopwords_path"]

        _io = _yaml_data["io"]
        self.root = _io["root"]
        self.save_dir = _io["save_dir"]
        self.vector_space_file = _io["vector_space_file"]
        self.ngram_vectors_file = _io["ngram_vectors_file"]
        self.word_map_file = _io["word_map_file"]

    def _get_save_dir_path(self) -> str:
        return os.path.join(self.root, self.save_dir)

    def get_vector_space_file_path(self) -> str:
        return os.path.join(self._get_save_dir_path(), self.vector_space_file)

    def get_ngram_vectors_file_path(self) -> str:
        return os.path.join(self._get_save_dir_path(), self.ngram_vectors_file)

    def get_word_map_file_path(self) -> str:
        return os.path.join(self._get_save_dir_path(), self.word_map_file)


class ClusteringConfig(Config):
    def __init__(self, clustering_model_name: str):
        super.__init__()
        self.data = self.load_yaml(
            os.path.join(self.config_dir, "clustering", clustering_model_name, "config.yaml")
        )


class GlobalConfig:
    def __init__(self, model_name: str, language: str) -> None:
        self.model_config: ModelConfig = ModelConfig(model_name)
        self.language_config: LanguageConfig = LanguageConfig(language)
